//**************************************************************************//
//                                                                          //
//     RC MODULE, LTD.                                                      //
//                                                                          //
//     NAME:             FHT (Fast Hadamard Transform)                      //
//                       256 points (2**8)                                  //
//                                                                          //
//     Revision Date:    17/03/99                                           //
//     Author:           Vitali Kashkarov                                   //
//                                                                          //
//     DATA                                                                 //
//           Input data:  8 bits (signed)                                   //
//           Result data: no more than 16 bits (signed)                     //
//     USAGE                                                                //
//           These are C-callable routines that implement:                  //
//               1. First 3 steps of FHT in parallel                        //
//                  with data conversion from 8 to 16 bits:                 //
//                                                                          //
//                    void Steps_1_3( void* ASource, void* AResult );       //
//                      ASource - array of 256 bytes.(on global bus)        //
//                      AResult - array of 256 shorts.(on local bus)        //
//                                                                          //
//               2. Next 5 steps of FHT in parallel.                        //
//                                                                          //
//                    void Steps_4_8( void* ASource, void* AResult );       //
//                      ASource - array of 256 shorts.(on local bus)        //
//                      AResult - array of 256 shorts.(on global bus)       //
//                                                                          //
//           The routines are called from C file Hadamard.c                 //
//           Their headers are defined in Hadam256.h                        //
//                                                                          //
//     CYCLES                                                               //
//          349 cycles (including C calling overhead)                       //
//**************************************************************************//

// Labels global description to access the routines from external C program
global _Steps_1_3: label;
global _Steps_4_8: label;

// Section to allocate memory on global bus
nobits "dataGLOBAL" 
    global _GBuff: long[40h];
end "dataGLOBAL";

// Section to allocate memory on local bus
nobits "dataLOCAL"  
    global _LBuff: long[40h];
end "dataLOCAL";

// Section of weights to load into the Active Matrix Unit of NM6403
data ".data"
    M_1_3: long[16] = ( 00001000100010001hl, 0FFFF0001FFFF0001hl,
                        0FFFFFFFF00010001hl, 00001FFFFFFFF0001hl,
                        00001000100010001hl, 0FFFF0001FFFF0001hl,
                        0FFFFFFFF00010001hl, 00001FFFFFFFF0001hl,
                        00001000100010001hl, 0FFFF0001FFFF0001hl,
                        0FFFFFFFF00010001hl, 00001FFFFFFFF0001hl,
                        0FFFFFFFFFFFFFFFFhl, 00001FFFF0001FFFFhl,
                        000010001FFFFFFFFhl, 0FFFF00010001FFFFhl );

    M_4_8: long[32] = ( 05555555555555555hl, 0DDDDDDDDDDDDDDDDhl,
                        0F5F5F5F5F5F5F5F5hl, 07D7D7D7D7D7D7D7Dhl,
                        0FF55FF55FF55FF55hl, 077DD77DD77DD77DDhl,
                        05FF55FF55FF55FF5hl, 0D77DD77DD77DD77Dhl,
                        0FFFF5555FFFF5555hl, 07777DDDD7777DDDDhl,
                        05F5FF5F55F5FF5F5hl, 0D7D77D7DD7D77D7Dhl,
                        055FFFF5555FFFF55hl, 0DD7777DDDD7777DDhl,
                        0F55F5FF5F55F5FF5hl, 07DD7D77D7DD7D77Dhl,
                        0FFFFFFFF55555555hl, 077777777DDDDDDDDhl,
                        05F5F5F5FF5F5F5F5hl, 0D7D7D7D77D7D7D7Dhl,
                        055FF55FFFF55FF55hl, 0DD77DD7777DD77DDhl,
                        0F55FF55F5FF55FF5hl, 07DD77DD7D77DD77Dhl,
                        05555FFFFFFFF5555hl, 0DDDD77777777DDDDhl,
                        0F5F55F5F5F5FF5F5hl, 07D7DD7D7D7D77D7Dhl,
                        0FF5555FF55FFFF55hl, 077DDDD77DD7777DDhl,
                        05FF5F55FF55F5FF5hl, 0D77D7DD77DD7D77Dhl);
end ".data";

// Section of source code.
begin ".text"    
// The routine for the first 3 steps of FHT.
<_Steps_1_3>

.branch; // switch on the parallel vector instructions execution.

    // Get position in stack to access the parametrs of the routine.
    ar5 = ar7 - 2;
    
    // Save registers 
    push ar0, gr0;
    push ar4, gr4;
    
    // Load the columns configuration of the Shadow Matrix.
    gr0 = 80008000h;

    push ar5, gr5;
    // Copy the columns configuration into nb1 | Init gr4 with zero.
    nb1 = gr0 with gr4 = false;

    // Load the rows configuration of the Shadow Matrix.
    sb  = 03030303h;

    // Load the address of weights array | gr4 = 1;
    ar0 = M_1_3 with gr4++;

    // Load 16 long words to wfifo, transfer 8 words to the Shadow Matrix
    // and copy the Shadow Matrix contents to the Active Matrix.
    rep 16 wfifo = [ar0++], ftw, wtw;

    // Load the address of source buffer | gr4 = 2;
    ar0 = [--ar5] with gr4 <<= 1;      
    // Load the address of destination buffer | gr5 = 4; <- Increment value
    ar4 = [--ar5] with gr5 = gr4 << 1;
    // ar5 points to the destination buffer address + 2 (next long word).
    ar5 = ar4 + gr4 with gr4 = gr5;
    
    //  Load source data, make calculations and 
    //  transfer 8 long words from wfifo to the Shadow Matrix.
    rep 32 ram = [ar0++],ftw with vsum , data, 0;
    //  Store results in memory. // These two instructions are executed
    rep 32 [ar4++gr4] = afifo;   // in parallel.

    // This part of code is used to step over the silicon bug.
.wait;    // switch off the parallel vector instructions execution.
    // Copy the same constant to nb1 to lock wtw execution 
    // until ftw is finished.
    nb1 = gr0;
    // Copy the Shadow Matrix contents to the Active Matrix.
    wtw;
.branch;  // switch on the parallel vector instructions execution.

    // Make a second part of calculations.
    rep 32 with vsum , ram, 0;   // These two instructions are executed
    // Store results in memory.  // in parallel.
    rep 32 [ar5++gr5] = afifo;   // 
                                 
    // Restore previous registers' values
    pop ar5, gr5;
    pop ar4, gr4;
    pop ar0, gr0;

    // Return from the routine.
    return;
.wait;


// The routine for the next 5 steps of FHT.
<_Steps_4_8>
.branch; // switch on the parallel vector instructions execution.
    
    // Get position in stack to access the parametrs of the routine.
    ar5 = ar7 - 2;
    
    // Save registers 
    push ar0, gr0;
    push ar1, gr1;
    push ar2, gr2;
    push ar4, gr4;
    push ar5, gr5;
    push ar6, gr6;

    // Load the address of weights array | Init gr0 with zero.
    ar6 = M_4_8 with gr0 = false;

    // Load the columns configuration of the Shadow Matrix.
    gr2 = 80008000h;

    // Load the rows configuration of the Shadow Matrix.
    sb  = 0FFFFFFFFh;

    // Copy the columns configuration into nb1 | gr0 = 1;
    nb1 = gr2 with gr0++;

    // Load the address of source buffer | gr0 = 2;
    ar0 = [--ar5] with gr0 <<= 1; 
    
    // ar1 points to the source buffer address + 2 (next long word)| gr1 = 4
    ar1 = ar0 + gr0 with gr1 = gr0 << 1;

    // Load the address of destination buffer | gr4 = 2;
    ar4 = [--ar5] with gr4 = gr0; 

    // ar1 points to the destination buffer address + 2 | gr4 = 4
    ar5 = ar4 + gr4 with gr5 = gr1;

    // gr0 = 4 | gr4 = 4 <- Increments to access odd/even long words.
    gr0 = gr1 with gr4 = gr1;

    // Load source data to wfifo -> ShM -> ActM
    rep 32 wfifo = [ar0++gr0], ftw, wtw;   
    // Load weights to ram;      // These two instructions are executed
    rep 32 ram = [ar6++];        // in parallel.


    // Load next part of source data to wfifo -> ShM | Make calculations.
    rep 32 wfifo = [ar1++gr1], ftw with vsum , ram, 0;
    // Store results in memory.  // These two instructions are executed 
    rep 32 [ar4++gr4] = afifo;   // in parallel.

    // This part of code is used to step over the silicon bug.
.wait;    // switch off the parallel vector instructions execution.
    // Copy the same constant to nb1 to lock wtw execution 
    // until ftw is finished.
    nb1 = gr2;
    // Copy the Shadow Matrix contents to the Active Matrix.
    wtw;
.branch;  // switch on the parallel vector instructions execution.

    // Make calculations.        //
    rep 32 with vsum , ram, 0;   // These two instructions are executed 
    // Store results in memory.  // in parallel.
    rep 32 [ar5++gr5] = afifo;   //

    // Restore previous registers' values
    pop ar6, gr6;
    pop ar5, gr5;
    pop ar4, gr4;
    pop ar2, gr2;
    pop ar1, gr1;
    pop ar0, gr0;

    // Return from the routine.
    return;
.wait;

end ".text";

//**************************************************************************//
//                              End of file                                 //
//**************************************************************************//
